local L = Capitalist.Locale
local realm = GetRealmName()
local character = UnitName("player")

local productsMemory = { }

function Capitalist:ProductMemoryReset()
	productsMemory = { }
end

-- REVIEW: Needs a better name
function Capitalist:SkillIsCyclic(profession, skillName)
	-- TODO: Arcanite Bar
	-- REVIEW: Exclude transmutes to avoid infinite recursion (earth -> fire -> earth ...)
	if (skillName:find(L["Transmute:"])) then
		return true
	end
	return false
end

function Capitalist:ProductIsEnabled(profession, skillName)
	return (self.db.profile.professions and
			self.db.profile.professions[profession] and
			self.db.profile.professions[profession][skillName] and
			self.db.profile.professions[profession][skillName].enabled)
end

function Capitalist:ProductSetEnabled(profession, skillName, enabled)
	if (enabled) then
		self.db.profile.professions[profession][skillName].enabled = true
	else
		self.db.profile.professions[profession][skillName].enabled = nil
	end
end

function Capitalist:ProductComponentsAreVisible(profession, skillName)
	return (not productsMemory[profession]) or
		   (not productsMemory[profession][skillName]) or
		   (not productsMemory[profession][skillName].hiddenComponents)
end

function Capitalist:ProductIsAllowed(profession, skillName)
	if ((not self.db.profile.professions) or
		(not self.db.profile.professions[profession]) or
		(not self.db.profile.professions[profession][skillName])) then
		return false
	end
	
	local skillData = self.db.profile.professions[profession][skillName]
	local recipeLink = skillData.recipeLink
		
	if ((not L.expansionRanges[profession]) or (not recipeLink)) then
		return true
	end

	local recipeIdString = string.match(recipeLink, "Henchant:(%d+)")
	local recipeIdNumber = tonumber(recipeIdString)

	if (recipeIdNumber >= L.expansionRanges[profession]["WOTLK"]) then
		return self.db.profile.filters.global.Northrend
	elseif (recipeIdNumber >= L.expansionRanges[profession]["BC"]) then
		return self.db.profile.filters.global.BurningCrusade
	else
		return self.db.profile.filters.global.Classic
	end
end

function Capitalist:ProductComponentsSetVisible(profession, skillName, visible)
	if (visible) then
		if (productsMemory[profession]) then
			if (productsMemory[profession][skillName]) then
				productsMemory[profession][skillName].hiddenComponents = nil
			end
		end
	else
		if (not productsMemory[profession]) then
			productsMemory[profession] = {}
		end
		if (not productsMemory[profession][skillName]) then
			productsMemory[profession][skillName] = {}
		end
		productsMemory[profession][skillName].hiddenComponents = true
	end
end

function Capitalist:ProductIsVisible(profession, skillName)
	return (not productsMemory[profession]) or
		   (not productsMemory[profession][skillName]) or
		   (not productsMemory[profession][skillName].hidden)
end

function Capitalist:ProductSetVisible(profession, skillName, visible)
	if (visible) then
		if (productsMemory[profession]) then
			if (productsMemory[profession][skillName]) then
				productsMemory[profession][skillName].hidden = nil
			end
		end
	else
		if (not productsMemory[profession]) then
			productsMemory[profession] = {}
		end
		if (not productsMemory[profession][skillName]) then
			productsMemory[profession][skillName] = {}
		end
		productsMemory[profession][skillName].hidden = true
	end
end

function Capitalist:FindProfession(skillName)
	if (self.db.profile.professions) then
		for profession, professionData in pairs(self.db.profile.professions) do
			if (professionData[skillName]) then
				return profession
			end
		end
	end
	return nil
end

function Capitalist:FindCrafter(profession, skillName)
	if (self.db.profile[realm]) then
		for crafter, crafterData in pairs(self.db.profile[realm]) do
			if (crafterData[profession] and crafterData[profession][skillName]) then
				return crafter
			end
		end
	end
	return nil
end

function Capitalist:IsProduct(productName)
	local skillName, profession = self:SkillNameFromProductName(productName)
	if (skillName and profession) then
		return true
	end
	return false
end

-- REVIEW: callers should really use SkillsFromProductName and react to multiple answers
-- REVIEW: expensive, cache in products memory?
function Capitalist:SkillNameFromProductName(productName)
	if (self.db.profile.professions) then
		for profession, professionData in pairs(self.db.profile.professions) do
			for skillName, skillData in pairs(professionData) do
				if (not self:SkillIsCyclic(profession, skillName)) then
					if (productName == self:ProductNameFromSkillName(profession, skillName)) then
						return skillName, profession
					end
				end
			end
		end
	end
	return nil, nil
end

function Capitalist:SkillsFromProductName(productName)
	local skills
	if (self.db.profile.professions) then
		for profession, professionData in pairs(self.db.profile.professions) do
			for skillName, skillData in pairs(professionData) do
				if (not self:SkillIsCyclic(profession, skillName)) then
					if (productName == self:ProductNameFromSkillName(profession, skillName)) then
						if (not skills) then
							skills = {}
						end
						skills[skillName] = profession
					end
				end
			end
		end
	end

	return skills
end

function Capitalist:ProductNameFromSkillName(profession, skillName)
	if (self.db.profile.professions[profession][skillName].itemLink) then
		local productName = self.db.profile.professions[profession][skillName].itemLink:match("%[(.*)%]")
		if (productName:find(L["Enchant "])) then
			productName = L["Scroll of "]..productName
		end

		return productName
	end

	return nil
end

function Capitalist:ProductTextFromSkillName(profession, skillName)
	if (self.db.profile.professions[profession][skillName].itemLink) then
		return self.db.profile.professions[profession][skillName].itemLink
	else
		return self:ProductNameFromSkillName(profession, skillName)
	end
end

function Capitalist:SkillTextFromSkillName(profession, skillName)
	if (self.db.profile.professions[profession][skillName].recipeLink) then
		return self.db.profile.professions[profession][skillName].recipeLink
	else
		return skillName
	end
end

local function ColorizeForSkillType(input, skillType)
	if (skillType == "trivial") then
		return "|cff808080"..input.."|r"
	elseif (skillType == "easy") then
		return "|cff00ff00"..input.."|r"
	elseif (skillType == "medium") then
		return "|cffffff00"..input.."|r"
	elseif (skillType == "optimal") then
		return "|cffff8000"..input.."|r"
	elseif (skillType == "difficult") then
		return "|cffff8080"..input.."|r"
	else
		return input
	end
end

function Capitalist:OutputProductLink(profession, prefix, skillName, count, crafter)
	if (self.db.profile.professions and
		self.db.profile.professions[profession] and
		self.db.profile.professions[profession][skillName]) then

		local itemText = self:ProductTextFromSkillName(profession, skillName)

		local s = self:StringForItem(prefix, itemText, count)
		if (self.db.profile.professions[profession][skillName].min) then
			s = s.." ("..self.db.profile.professions[profession][skillName].min
			if (self.db.profile.professions[profession][skillName].max) then
				s = s.."-"..self.db.profile.professions[profession][skillName].max
			end
			s = s..")"
		end

		local cost = self:CalculateSkillCost(profession, skillName)
		if (cost) then
			s = s.." @ "..self:PrettyCurrency(cost).." ea."
		end		

		if (crafter) then
			local skillType = self.db.profile[realm][crafter][profession][skillName].type
			s = s.." by "..ColorizeForSkillType(crafter, skillType)
		end
		s = s.." with "..self:SkillTextFromSkillName(profession, skillName)
		
		self:Print(s)
	end
end

function Capitalist:FindProducts(...)
	local pattern = ...
	if (self.db.profile.professions) then
		for profession, professionData in pairs(self.db.profile.professions) do
			for skillName, skillData in pairs(professionData) do
				if (strlower(skillName):find(strlower(pattern))) then
					self:OutputProductLink(profession, "", skillName, 1)
				end
			end
		end
	end
end

function Capitalist:OutputProductCraftingSequence(...)
	local productName, count, prefix = ...
	if (not count) then
		count = 1
	end
	if (not prefix) then
		prefix = ""
	end

	local skills = self:SkillsFromProductName(productName)
	if (skills) then
		for skillName, profession in pairs(skills) do
			local crafter = self:FindCrafter(profession, skillName)
			self:OutputProductLink(profession, prefix, skillName, count, crafter)

			if (crafter) then
				for componentName, componentCount in pairs(self.db.profile.professions[profession][skillName].components) do
					self:OutputProductCraftingSequence(componentName, componentCount, prefix.."    ")
				end
			end
		end
	else
		self:OutputComponent(prefix, productName, count)
	end
end

function Capitalist:ReportTop10(...)
	self:DebugEntry("ReportTop10")

	local tableSorted = {}

	if (self.db.profile[realm] and
		self.db.profile[realm][character]) then

		for profession, professionData in pairs(self.db.profile[realm][character]) do
			for skillName, skillData in pairs(professionData) do
				local numCreated = skillData.numCreated or 0
				if (numCreated > 0) then
					local entry = { c = numCreated, n = skillName }
					table.insert(tableSorted, entry )
				end
			end
		end
	end

	table.sort(tableSorted, function(a, b) return a.c > b.c end)

	for i=1, math.min(#tableSorted, 10) do
		local e = tableSorted[i]
		self:Print(e.c.."x "..e.n)
	end

	self:DebugExit("ReportTop10")
end
